/* 
 * Cross Platform Mutex abstraction
 * Copyright(C) 2007 Michael Jerris
 *
 * You may opt to use, copy, modify, merge, publish, distribute and/or sell
 * copies of the Software, and permit persons to whom the Software is
 * furnished to do so.
 *
 * This work is provided under this license on an "as is" basis, without warranty of any kind,
 * either expressed or implied, including, without limitation, warranties that the covered code
 * is free of defects, merchantable, fit for a particular purpose or non-infringing. The entire
 * risk as to the quality and performance of the covered code is with you. Should any covered
 * code prove defective in any respect, you (not the initial developer or any other contributor)
 * assume the cost of any necessary servicing, repair or correction. This disclaimer of warranty
 * constitutes an essential part of this license. No use of any covered code is authorized hereunder
 * except under this disclaimer. 
 *
 * Contributors: 
 *
 *
 */

#ifdef WIN32
#   if (_WIN32_WINNT < 0x0400)
#       error "Need to target at least Windows 95/WINNT 4.0 because TryEnterCriticalSection is needed"
#   endif
#   include <windows.h>
#endif

#ifndef SCCI_DEBUG_MUTEX
#define SCCI_DEBUG_MUTEX
#endif

#include "scci_mutex.h"
#include "scci_thread.h"
#include "scci_memory.h"
#include "scci_log.h"
				 

#ifdef WIN32
#include <process.h>

#define SCCI_THREAD_CALLING_CONVENTION __stdcall

struct scci_mutex {
	CRITICAL_SECTION mutex;
};

#else
#include <pthread.h>
#include <poll.h>

#include <errno.h>
#ifndef PTHREAD_MUTEX_RECURSIVE
#define PTHREAD_MUTEX_RECURSIVE PTHREAD_MUTEX_RECURSIVE_NP
#endif /* PTHREAD_MUTEX_RECURSIVE */
				  
#ifdef SCCI_DEBUG_MUTEX
#define SCCI_MUTEX_MAX_REENTRANCY 30
typedef struct scci_lock_entry {
	const char *file;
	const char *func;
	uint32_t line;
} scci_lock_entry_t;

typedef struct scci_lock_history {
	scci_lock_entry_t locked;
	scci_lock_entry_t unlocked;
} scci_lock_history_t;
#endif

struct scci_mutex {
	pthread_mutex_t mutex;
#ifdef SCCI_DEBUG_MUTEX
	scci_lock_history_t lock_history[SCCI_MUTEX_MAX_REENTRANCY];
	uint8_t reentrancy;
#endif
};
#endif

SCCI_DECLARE(scci_status_t) scci_mutex_create(scci_mutex_t **mutex)
{
	scci_status_t status = SCCI_FAIL;
	scci_mutex_t *check = NULL;

	check = (scci_mutex_t *)scci_calloc(1, sizeof(scci_mutex_t));
	if (!check)
		goto done;
#ifdef WIN32
	InitializeCriticalSection(&check->mutex);
#else
	if (pthread_mutex_init(&check->mutex, NULL))
		goto fail;

	goto success;

 fail:
	goto done;

 success:
#endif
	*mutex = check;
	status = SCCI_SUCCESS;

 done:
	return status;
}

SCCI_DECLARE(scci_status_t) scci_mutex_destroy(scci_mutex_t **mutex)
{
	scci_mutex_t *mp = *mutex;
	*mutex = NULL;
	if (!mp) {
		return SCCI_FAIL;
	}
#ifdef WIN32
	DeleteCriticalSection(&mp->mutex);
#else
	if (pthread_mutex_destroy(&mp->mutex))
		return SCCI_FAIL;
#endif
	scci_safe_free(mp);
	return SCCI_SUCCESS;
}

#define ADD_LOCK_HISTORY(mutex, file, line, func) \
	{ \
		if ((mutex)->reentrancy < SCCI_MUTEX_MAX_REENTRANCY ) { \
			(mutex)->lock_history[mutex->reentrancy].locked.file = (file); \
			(mutex)->lock_history[mutex->reentrancy].locked.func = (func); \
			(mutex)->lock_history[mutex->reentrancy].locked.line = (line); \
			(mutex)->lock_history[mutex->reentrancy].unlocked.file = NULL; \
			(mutex)->lock_history[mutex->reentrancy].unlocked.func = NULL; \
			(mutex)->lock_history[mutex->reentrancy].unlocked.line = 0; \
			(mutex)->reentrancy++; \
			if ((mutex)->reentrancy == SCCI_MUTEX_MAX_REENTRANCY) { \
				scci_el_default_logger(file, func, line, SCCI_LOG_ERROR, "Max reentrancy reached for mutex %p (%s:%s:%d)\n", (mutex), (file), (func), (line)); \
			} \
		} \
	}

SCCI_DECLARE(scci_status_t) _scci_mutex_lock(const char *file, int line, const char *func, scci_mutex_t *mutex)
{
#ifdef WIN32
	UNREFERENCED_PARAMETER(file);
	UNREFERENCED_PARAMETER(line);
	UNREFERENCED_PARAMETER(func);

	EnterCriticalSection(&mutex->mutex);
#else
	int err;
	if ((err = pthread_mutex_lock(&mutex->mutex))) {
		scci_el_default_logger(file, __FUNCTION__, line, SCCI_LOG_ERROR, "Failed to lock mutex %d\n", err);
		return SCCI_FAIL;
	}
#endif
#ifdef SCCI_DEBUG_MUTEX 
	ADD_LOCK_HISTORY(mutex, file, line, func);
#endif
	return SCCI_SUCCESS;
}

SCCI_DECLARE(scci_status_t) _scci_mutex_unlock(const char *file, int line, const char *func, scci_mutex_t *mutex)
{
#ifdef SCCI_DEBUG_MUTEX
	int i = 0;
	if (mutex->reentrancy == 0) {
		scci_el_default_logger(file, __FUNCTION__, line, SCCI_LOG_ERROR, "Cannot unlock something that is not locked! (%s:%s:%d)\n", file, func, line);
		return SCCI_FAIL;
	}
	i = mutex->reentrancy - 1;
	/* I think this is a fair assumption when debugging */
	if (func != mutex->lock_history[i].locked.func) {
		scci_el_default_logger(file, __FUNCTION__, line,SCCI_LOG_WARNING, "Mutex %p was suspiciously locked at %s->%s:%d but unlocked (%s:%s:%d)\n",
				mutex, mutex->lock_history[i].locked.func, mutex->lock_history[i].locked.file, mutex->lock_history[i].locked.line,
				func, file, line);
	}
	mutex->lock_history[i].unlocked.file = file;
	mutex->lock_history[i].unlocked.line = line;
	mutex->lock_history[i].unlocked.func = func;
	mutex->reentrancy--;
#endif
#ifdef WIN32
	UNREFERENCED_PARAMETER(file);
	UNREFERENCED_PARAMETER(line);
	UNREFERENCED_PARAMETER(func);

	LeaveCriticalSection(&mutex->mutex);
#else
	if (pthread_mutex_unlock(&mutex->mutex)) {
		scci_el_default_logger(file, __FUNCTION__, line,SCCI_LOG_ERROR, "Failed to unlock mutex: %d\n", errno);
#ifdef SCCI_DEBUG_MUTEX
		mutex->reentrancy++;
#endif
		return SCCI_FAIL;
	}
#endif
	return SCCI_SUCCESS;
}

/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4:
 */
