/*
 * Copyright (C) 2011, Sangoma Technologies.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contributors:
 *
 */

#include "scci_thread.h"
#include "scci_memory.h"
#include "scci_def.h"
#include "scci_mutex.h"
#include "scci_queue.h"

scci_status_t scci_queue_create (scci_queue_t **outqueue, scci_size_t capacity)
{
	scci_queue_t *queue = NULL;

	scci_assert_return(outqueue, SCCI_FAIL, "Queue double pointer is null\n");
	scci_assert_return(capacity > 0 , SCCI_FAIL, "Queue capacity is not bigger than 0\n");

	*outqueue = NULL;
	queue = scci_calloc(1, sizeof(*queue));
	scci_assert_return(queue, SCCI_FAIL, "Failed to alloc mem\n");

	queue->elements = scci_calloc(1, sizeof(void*)*capacity);
	if (!queue->elements) {
		goto failed;
	}
	queue->capacity = capacity;

	if (scci_mutex_create(&queue->mutex) != SCCI_SUCCESS) {
		goto failed;
	}

	*outqueue = queue;
	return SCCI_SUCCESS;

failed:
	if (queue) {
		scci_mutex_destroy(&queue->mutex);
		scci_safe_free(queue->elements);
		scci_safe_free(queue);
	}
	
	return SCCI_FAIL;
}

scci_status_t scci_queue_enqueue(scci_queue_t *queue, void *obj)
{
	scci_status_t status = SCCI_FAIL;
	
	scci_assert_return(queue, SCCI_FAIL, "Queue is null\n");

	scci_mutex_lock(queue->mutex);

	if (queue->windex == queue->capacity) {
		/* try to see if we can wrap around */
		queue->windex = 0;
	}

	if (queue->size != 0 && queue->windex == queue->rindex) {
		scci_el_default_logger(__FILE__, __FUNCTION__, __LINE__,SCCI_LOG_WARNING, "Failed to enqueue obj %p in queue %p, no more room! windex == rindex == %d!\n", obj, queue, queue->windex);
		goto done;
	}

	queue->elements[queue->windex++] = obj;
	queue->size++;
	status = SCCI_SUCCESS;

done:
	scci_mutex_unlock(queue->mutex);

	return status;
}

scci_bool_t scci_queue_empty(scci_queue_t *queue)
{
	return (queue->size) ? SCCI_FALSE : SCCI_TRUE;
}

void *scci_queue_dequeue(scci_queue_t *queue)
{
	void *obj = NULL;

	scci_assert_return(queue, NULL, "Queue is null!");
	scci_mutex_lock(queue->mutex);

	if (queue->size == 0) {
		goto done;
	}

	obj = queue->elements[queue->rindex];
	queue->elements[queue->rindex++] = NULL;
	queue->size--;
	if (queue->rindex == queue->capacity) {
		queue->rindex = 0;
	}

done:
	scci_mutex_unlock(queue->mutex);

	return obj;
}

scci_status_t scci_queue_destroy(scci_queue_t **inqueue)
{
	scci_queue_t *queue = NULL;
	scci_assert_return(inqueue, SCCI_FAIL, "Queue is null!\n");
	scci_assert_return(*inqueue, SCCI_FAIL, "Queue is null!\n");

	queue = *inqueue;
	scci_mutex_destroy(&queue->mutex);
	scci_safe_free(queue->elements);
	scci_safe_free(queue);
	*inqueue = NULL;

	return SCCI_SUCCESS;
}
